/* $Id: misc.c,v 1.100 2004/01/22 15:44:26 rav Exp $ */

/* Intro {{{
 * ---------------------------------------------------------------- 
 *  DConnect Daemon
 * 
 *      #(@) Copyright (c) 2002, DConnect development team
 *      #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 *  This  program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You  should  have  received a copy  of  the GNU General Public
 *  License  along  with  this  program;  if  not,  write  to  the
 *  Free  Software  Foundation,  Inc.,  675  Mass  Ave,  Cambridge,
 *  MA 02139, USA. 
 *
 * ---------------------------------------------------------------- 
 * }}} */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "match.h"
#include "conf.h"
#include "net.h"
/* }}} */

extern struct userrec *user[];
extern struct config conf;
extern int NUSERS;
extern struct request_info req;
extern char *__progname;
extern const char *reserved_nicks[];
extern FILE* file_net;
extern FILE* file_proto;
extern FILE* file_wrng;
extern FILE* file_con;
extern FILE* file_chat;
extern char DAEMON;

/* controlled free() {{{ */
void my_free(void *ptr) {
    if (ptr) free(ptr);
    ptr=NULL;
} /* }}} */

/* controlled realloc() {{{ */
void *my_realloc(void *ptr,size_t size) {
    void *tmp;
    if (size>MAXBLOCK) {
	fprintf(stderr,"Attempt to allocate %dkB of data - refused.\r\n",size/1024);
	my_free(ptr);
	return NULL;
    }
    tmp=realloc(ptr,size);
    if (tmp==NULL) {
	fprintf(stderr,"FATAL: Memory allocation failed.\r\n");
	exit(0);
    }
    return tmp;
} /* }}} */

/* controlled malloc() {{{ */
void *my_malloc(size_t size) {
    return my_realloc(NULL,size);
} /* }}} */

/* show_usage() - just show a brief usage {{{ */
void show_usage( void ) {
	printf( "DConnect Daemon v%s\r\n\r\n",PACKAGE_VERSION );
	printf( "Usage:\t%s [options]\r\n\r\n", __progname );
	printf( "-h\t\t this help screen\r\n" );
	printf( "-v\t\t dump version and exit\r\n" );
	printf( "-c config_file\t use alternate config file\r\n" );
	printf( "\r\nReport bugs to <dc-hub@ds.pg.gda.pl>.\r\n" );
	exit( 0 );
} /* }}} */

/* parse_options() - parse command-line options {{{ */
void parse_options( struct config *conf, int argc, char * argv[] ) {
	int i;

	for ( i = 1; i < argc; i++ )
	{
		if ( argv[i][0] != '-' ) 
			show_usage( );
		
		/* There should be no more options. Use config file for the rest. */
		switch(argv[i][1]) {
			case 'h': show_usage();
			case 'v': printf("%s\r\n",PACKAGE_VERSION);
				  exit(0);
			case 'c': if (argc<i+2) show_usage(); else i++;
				  my_free(conf->conf_main);
				  conf->conf_main=strdup(argv[i]);
				  break;
			default: fprintf(stderr,"Unknown option: `%s'. Try `-h' for help.\r\n\r\n",argv[i]);
				 exit(0);
		}
	}
} /* }}} */

/* get_progname() - from openssh {{{ */
char *get_progname(char *argv0) {
#ifdef HAVE___PROGNAME
	extern char *__progname;
	return __progname;
#else
	char *p;
	if (argv0==NULL) return "unknown";
	
	p=strrchr(argv0,'/');
	if (p==NULL) p=argv0; else p++;
	return p;
#endif
} /* }}} */

/* randomize()	- feed a random numbers generator {{{ */
void randomize( void ) {
	unsigned int seed=0;
	
	FILE * rdev = fopen( "/dev/urandom", "r" );
	if ( rdev != NULL ) {
		seed = fgetc( rdev ); fclose( rdev );
	}
 
	seed += getpid() + time( NULL );
	srand( seed );
} /* }}} */

/* lockgen() - lock generator: 'salt Pk=privkey' {{{ */
char *lockgen() {
	char *lock;
	int salt_len, pk_len, n;
	
	salt_len = 50 + rand() % 50;
	pk_len = 5;
	lock=(char *)my_malloc(salt_len+pk_len+5);	
	for(n=0;n<salt_len;n++) lock[n]='%'+rand()%('z'-'%');
	lock[salt_len]=0;
	strcat(lock," Pk=");
	for(n=0;n<pk_len;n++) lock[salt_len+4+n]='%'+rand()%('z'-'%');
	lock[salt_len+4+pk_len]=0;
	return lock;
} /* }}} */

/* make_string() - allocate a sufficiently large string and print into it {{{ */
/* taken from man pages */
char *make_string(const char *fmt,...) {
	int n,size=128;
	char *p;
	va_list ap;

	if ((p=my_malloc(size))==NULL) return NULL;
	while (1) {
                va_start(ap,fmt);
		n=vsnprintf(p,size,fmt,ap);
                va_end(ap);
		if (n>-1&&n<size) return p;
		/* try again with more space. */
		if (n>-1) size=n+1;	/* glibc 2.1 */
		else size*=2;		/* glibc 2.0 */
		if ((p=my_realloc(p,size))==NULL) return NULL;
	}
} /* }}} */

/* make_message() - allocate a sufficiently large string and print into it {{{ */
/* simplified version */
char *make_message(const char *fmt,va_list ap) {
	int n,size=128;
	char *p;
	
	if ((p=my_malloc(size))==NULL) return NULL;
	while (1) {
		n=vsnprintf(p,size,fmt,ap);
		if (n>-1&&n<size) return p;
		/* try again with more space. */
		if (n>-1) size=n+1;	/* glibc 2.1 */
		else size*=2;		/* glibc 2.0 */
		if ((p=my_realloc(p,size))==NULL) return NULL;
	}
} /* }}} */

/* debug() - debugging / logging {{{ */
void debug(char level,const char *fmt,...) {
	int n;
	char *str,timer[50];
	unsigned char lvl;
	time_t tm;
	va_list args;

	/* compose the debug message */
	va_start(args,fmt);
	str=(char *)make_message(fmt,args);
	va_end(args);

	lvl=conf.log_level&level;

	time(&tm);
	strftime(timer,sizeof(timer),"%Y.%m.%d %T",localtime(&tm));

	/* print to syslog */
	if (lvl&DEBUG_SYS) {
	    openlog("dcd",LOG_PID|LOG_CONS,LOG_DAEMON);
	    syslog(LOG_NOTICE,"%s\n",str);
	}

	/* print to main log or console */
	if (lvl&DEBUG_STD) {
	    if (DAEMON) fprintf(stderr,"[%s] ",timer);
	    fprintf(stderr,"%s\r\n",str);
	    fflush(stderr);
	}

	/* print to net log */
	if (file_net) if (lvl&DEBUG_NET) {
	    fprintf(file_net,"[%s] ",timer);
	    fprintf(file_net,"%s\r\n",str);
	    fflush(file_net);
	}

	/* print to proto log */
	if (file_proto) if (lvl&DEBUG_PROTO) {
	    fprintf(file_proto,"[%s] ",timer);
	    fprintf(file_proto,"%s\r\n",str);
	    fflush(file_proto);
	}

	/* print to wrng log */
	if (file_wrng) if (lvl&DEBUG_WRNG) {
	    fprintf(file_wrng,"[%s] ",timer);
	    fprintf(file_wrng,"%s\r\n",str);
	    fflush(file_wrng);
	}

	/* print to console log */
	if (file_con) if (lvl&DEBUG_CON) {
	    fprintf(file_con,"[%s] ",timer);
	    fprintf(file_con,"%s\r\n",str);
	    fflush(file_con);
	}

	/* print to chat log */
	if (file_chat) if (lvl&DEBUG_CHAT) {
	    fprintf(file_chat,"[%s] ",timer);
	    fprintf(file_chat,"%s\r\n",str);
	    fflush(file_chat);
	}

	/* send message to console users */
	for (n=0;n<NUSERS;n++)
	    if (user[n] && user_tst_state(user[n],STATE_LOGGEDIN|STATE_REGISTERED)&& (user[n]->cons&level))
		sendf(user[n],"%s\r\n",str);

	my_free(str);
} /* }}} */

/* validnick() - check if a given nick is a valid one {{{ */
char validnick(struct userrec *usr,char *nick) {
	char *tmp=NULL;
	int n=0;
	int zwrot=0;

	if (!nick) return 0;
	if (!*nick) return 0;

	if (strlen(nick)+1>NICK_LEN) {
		publicmsg(usr,"Your nick is too long.");
		debug(DEBUG_PROTO,"Nick too long: %s@%s.",nick,usr->ip);
		return 0;
	}

        if (usr->cons==0 && validlogin(usr,nick,NULL)) 
	    {
		publicmsg(usr,"Your nick requiers password");
		
		my_free(usr->nick);
		usr->nick=strdup(nick);

		return 2;
	    }

	if ((usr->cons==0)&&(*(conf.conf_nallow)))
		if (!nick_allowed(usr,nick)) return 0;

	while(*reserved_nicks[n])
		if (!strcasecmp(nick,reserved_nicks[n++])) { 
			publicmsg(usr,"Your nick is reserved.");
			return 0; 
		}


	for(n=0;n<strlen(nick);n++)
		if (iscntrl(nick[n])||isspace(nick[n])||(nick[n]=='$')) {
			publicmsg(usr,"Your nick is invalid.");
			debug(DEBUG_PROTO,"Nick invalid: %s@%s.",nick,usr->ip);
			return 0; 
		}

	for(n=0;n<NUSERS;n++)
		if (user[n] && 
		    user_tst_state(user[n],STATE_REGISTERED|STATE_LOGGEDIN) && 
		    !strcmp(user[n]->nick,nick)) 
		{
			publicmsg(usr,(usr->cons)?"Login incorrect.":"Your nick is already in use.");
			return 0;
		}

	my_free(tmp);
	tmp=strdup(nick);
	my_free(usr->nick);
	usr->nick=tmp;

	return 1;
} /* }}} */

/* validkey() - check for a valid $Key sequence - FIXME {{{ */
char validkey(char *key) {
	if (conf.allow_broken_key) return 1;
	if (!key) return 0;
	if (!*key) return 0;
	/* checking here - FIXME */
	return 1;
} /* }}} */

/* fill_info() - fill user's record with data received in $MyINFO {{{ */
char fill_info(struct userrec *usr,char *info) {
	char *i,*tmp,*bak;

	/* backup the string, 'cause it'll be changed after parsing */
	/* we also don't use strtok()s to leave it's stack for the
	 * code outside */
	bak=strdup(info);

	/* word 1 */
	tmp=strchr(bak,' '); 
	if (!tmp) goto error;

	/* word 2 */
	tmp++;
	i=strchr(tmp,' ');
	if (!i) goto error;
	*i=0;
	if (strcasecmp(tmp,"$ALL")) goto error;

	/* word 3 (nick) */
	tmp=i+1;
	i=strchr(tmp,' ');
	if (!i) goto error;
	*i=0;
	if (strcasecmp(tmp,usr->nick)) goto error;

	/* word 4 (description) */
	tmp=i+1;
	i=strchr(tmp,'$');
	if (!i) goto error;
	*i=0;
	if (strlen(tmp)+1>DESC_LEN) goto error;
	usr->desc=strdup(tmp);

	/* word 5 */
	tmp=i+1;
	i=strchr(tmp,'$');
	if (!i) goto error;

	/* word 6 (connection type) */
	tmp=i+1;
	i=strchr(tmp,'$');
	if (!i) goto error;
	*i=0;
	usr->flag=*(i-1);
	*(i-1)=0;
	if (strlen(tmp)+1>CTYPE_LEN) goto error;
	usr->ctype=strdup(tmp);
	
	/* word 7 (e-mail) */
	tmp=i+1;
	i=strchr(tmp,'$');
	if (!i) goto error;
	*i=0;
	if (strlen(tmp)+1>EMAIL_LEN) goto error;
	usr->email=strdup(tmp);
	
	/* word 8 (share bytes) */
	tmp=i+1;
	i=strchr(tmp,'$');
	if (!i) goto error;
	*i=0;
	sscanf(tmp,"%Lu",&(usr->share));

	/* parsing succeeded */
	my_free(bak);
	return 1;

	/* parsing error */
	error:
		 debug(DEBUG_PROTO,"On MyINFO from %s@%s: Error while parsing. (%s)",usr->nick,usr->ip,info);
		 my_free(bak);
		 return 0;
} /* }}} */

/* nick2id() - convert user's nick to table index {{{ */
int nick2id(char *nick) {
	int n;
	for(n=0;n<NUSERS;n++)
		if (user[n] && user_tst_state(user[n],STATE_LOGGEDIN|STATE_REGISTERED))
			if (!strcasecmp(user[n]->nick,nick)) return n;
	return -1;
} /* }}} */

/* usr2id() - convert user's data pointer to table index {{{ */
int usr2id(struct userrec *usr) {
	int n;
	for(n=0;n<NUSERS;n++) if (user[n]==usr) return n;
	debug(DEBUG_WRNG,"Internal error. usr2id() search failed.");
	return -1;
} /* }}} */

/* check_timeout() - check clients' times {{{ */
void check_timeout(struct userrec *usr) {
	time_t tv;

	if (!conf.ping_timeout&&(!conf.register_timeout|| (user_tst_state(usr,STATE_LOGGEDIN)))) return;
	
	time(&tv);
	if (conf.ping_timeout&&(usr->cons==0))
		if (tv-usr->ping>conf.ping_timeout) {
			deluser(usr,"Ping timeout.");
			return;
		}
	if (conf.register_timeout&&(user_tst_state(usr,STATE_CONNECTED|STATE_KEY|STATE_PASSWORD)))
		if (tv-usr->ping>conf.register_timeout) {
			deluser(usr,"Registration timeout.");
			return; 
		}
} /* }}} */

/* Buggy RH libwrap.a compilation has references to function
   yp_get_default_domain() in hosts_access(), but has no original
   function code linked. This is small own implemetation, that
   allows to compile and work properly on RedHat Linux */
/* yp_get_default_domain() - workaround for buggy RH libwrap {{{ */
int yp_get_default_domain(char **ptr) {
#define MAXHOSTNAMELEN 256
    static char mydomain[MAXHOSTNAMELEN];

    *ptr=mydomain;
    return (getdomainname(mydomain,MAXHOSTNAMELEN));
} /* }}} */

/* denied() - check if remote host is not allowed to connect {{{ */
char denied(int *sock) {
	request_init(&req,RQ_DAEMON,__progname,RQ_FILE,*sock,0);
	fromhost(&req);
	if (hosts_access(&req)) return 0;
	return 1;
} /* }}} */

/* make_path() - compose direct path to file {{{ */
void make_path(char *dir,char **file) {
	char *tmp,*p;

	if (!**file) return;
	tmp=strdup(*file);
	p=strrchr(tmp,'/');
	if (p) p++; else p=tmp;
	*file=(char *)my_realloc(*file,strlen(dir)+strlen(p)+2);
	sprintf(*file,"%s%s%s",dir,(dir[strlen(dir)-1]=='/')?"":"/",p);
	my_free(tmp);
} /* }}} */

/* make_paths() - compose direct paths to files for further use {{{ */
void make_paths(struct config *conf) {
	/* path to main log file */
	make_path(conf->log_dir,&(conf->log_main));

	/* path to net log file */
	make_path(conf->log_dir,&(conf->log_net));

	/* path to proto log file */
	make_path(conf->log_dir,&(conf->log_proto));

	/* path to wrng log file */
	make_path(conf->log_dir,&(conf->log_wrng));

	/* path to console log file */
	make_path(conf->log_dir,&(conf->log_con));

	/* path to chat log file */
	make_path(conf->log_dir,&(conf->log_chat));

	/* path to welcome message file */
	make_path(CONFIGDIR,&(conf->conf_welcome));

	/* path to motd file */
	make_path(CONFIGDIR,&(conf->conf_motd));

	/* path to bans file */
	make_path(CONFIGDIR,&(conf->conf_banned));

	/* path to console users file */
	make_path(CONFIGDIR,&(conf->conf_cusers));

	/* path to console host access file */
	make_path(CONFIGDIR,&(conf->conf_callow));

	/* path to allowed nicks file */
	make_path(CONFIGDIR,&(conf->conf_nallow));
} /* }}} */

/* uptime() - return string representing uptime {{{ */
char *uptime(time_t starttime) {
	static char str[32];
	time_t up;
	long d;
	char h,m,s;

	time(&up);
	up-=starttime;
	if (up<0) up=0;
	d=up/(3600*24);
	up-=d*3600*24;
	h=up/3600;
	up-=h*3600;
	m=up/60;
	up-=m*60;
	s=up;
	if (d) snprintf(str,sizeof(str),"%ld day%s, %s%d:%s%d:%s%d",d,(d==1)?"":"s",(h>9)?"":"0",h,(m>9)?"":"0",m,(s>9)?"":"0",s);
	else snprintf(str,sizeof(str),"%s%d:%s%d:%s%d",(h>9)?"":"0",h,(m>9)?"":"0",m,(s>9)?"":"0",s);
	return str;
} /* }}} */

/* toggle_flag() - toggle configuration flag {{{ */
inline void toggle_flag(struct config *conf,struct userrec *usr,unsigned char flag) {
    if (conf->flags&flag) conf->flags&=~flag; else conf->flags|=flag;
    sendf(usr,"Option has been turned %s.\r\n",(conf->flags&flag)?"on":"off");
} /* }}} */

#ifdef INCLUDED_DAEMON
/* daemon() - fall into background {{{ */
int daemon(int nochdir,int noclose) {
    pid_t ret;
#ifdef TIOCNOTTY
    int fd;
#endif

    ret=fork();
    if (ret) exit(0);
    if (ret<0) return (int)ret;
    if (noclose==0) {
        freopen("/dev/null", "r", stdin);
        freopen("/dev/null", "w", stdout);
        freopen("/dev/null", "w", stderr);
#ifdef TIOCNOTTY
        /* disconnect from the controlling tty */
        fd=open("/dev/tty",O_RDWR|O_NOCTTY);
        if (fd>=0) {
	    (void)ioctl(fd,TIOCNOTTY,NULL);
	    close(fd);
	}
#endif
    }
#ifdef HAVE_SETSID
    setsid();
#endif
    if (nochdir==0) chdir("/");
    return 0;
} /* }}} */
#endif /* INCLUDED_DAEMON */

/* VIM Settings {{{
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * soft-stop-width: 4
 * c indent on
 * End:
 * vim600: sw=4 ts=4 sts=4 cindent fdm=marker
 * vim<600: sw=4 ts=4
 * }}} */

